#!/usr/bin/env python
# coding:utf-8
import random, sys
import os.path
import io
import torch
import torch.distributed as dist
from PIL import Image
import torchvision
import imageio
from functools import partial

from aiges.core.types import *

try:
    from aiges_embed import ResponseData, Response, DataListNode, DataListCls, callback  # c++
except:
    from aiges.dto import Response, ResponseData, DataListNode, DataListCls, callback

from aiges.sdk import WrapperBase, JsonBodyField, StringBodyField, ImageBodyField, StringParamField
from aiges.utils.log import log, getFileLogger

import ailab.inference_wrapper.huggingface.diffusers.video_diffusion.wrapper.wan as wan
from ailab.inference_wrapper.huggingface.diffusers.video_diffusion.wrapper.wan.configs import \
    WAN_CONFIGS, SIZE_CONFIGS, MAX_AREA_CONFIGS, SUPPORTED_SIZES
from ailab.inference_wrapper.huggingface.diffusers.video_diffusion.wrapper.wan.modules.model import WanModel
from ailab.inference_wrapper.huggingface.diffusers.video_diffusion.wrapper.wan.modules.vae import WanVAE
from ailab.inference_wrapper.huggingface.diffusers.video_diffusion.wrapper.wan.modules.t5 import T5EncoderModel
from ailab.inference_wrapper.huggingface.diffusers.video_diffusion.wrapper.wan.modules.clip import CLIPModel
from ailab.inference_wrapper.huggingface.diffusers.video_diffusion.wrapper.wan.distributed.fsdp import shard_model

MODEL_NAME_MAPPING = {
    "wan21_t2v_1_3b": "t2v-1.3B",
    "wan21_t2v_14b": "t2v-14B",
    "wan21_i2v_14b_480p": "i2v-14B",
    "wan21_i2v_14b_720p": "i2v-14B"
}

# tmp_savefile = '/tmp/output_video.mp4'

# 定义服务推理逻辑
class Wrapper(WrapperBase):
    # serviceId = task
    version = "v1"
    # requestCls = UserRequest()
    # responseCls = UserResponse()
    model = None

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        import logging
        self.filelogger = getFileLogger(level=logging.INFO)
        self.sample_shift = 5.0
        self.base_seed = random.randint(0, sys.maxsize)
        self.sample_solver = 'unipc'
        self.sample_guide_scale = 5.0
        self.ulysses_size = 1
        self.ring_size = 1
        self.t5_fsdp = False
        self.dit_fsdp = False
        self.t5_cpu = True
        self.offload_model = False
        self.image_dir = None
        self.use_prompt_extend = False
        self.prompt_extend_method = 'local_qwen'
        self.prompt_extend_model = None
        self.prompt_extend_target_lang = 'cn'
        self.model = None
        self.vae = None
        self.text_encoder = None
        self.shard_fn = None
        self.clip = None

    def get_payload_text(self, reqData: DataListCls):
        text = reqData.get('video_description').data.decode('utf-8')
        return text
    
    def get_image(self, reqData: DataListCls):
        try:
            input = reqData.get("image").data
            img = Image.open(io.BytesIO(input))
            img = img.convert("RGB")
            return img
        except Exception as e:
            self.filelogger.error(f"Loading image error: {e}")
            return None
    
    def wrapperInit(self, config: {}) -> int:
        self.filelogger.info("Initializing ...")
        self.base_model = os.environ.get("FULL_MODEL_PATH")
        self.pretrained_name = os.environ.get("PRETRAINED_MODEL_NAME")
        self.task = MODEL_NAME_MAPPING[self.pretrained_name]

        self.filelogger.info(f"base_model:{self.base_model}")
        if not self.pretrained_name:
            self.filelogger.error("should have environ(PRETRAINED_MODEL_NAME,FULL_MODEL_PATH(lora weight dir))")
            return -1

        if not os.path.isdir(self.base_model):
            self.filelogger.error(f"not find the base_model in FULL_MODEL_PATH")
            return -1
        
        self.cfg = WAN_CONFIGS[self.task]
        self.filelogger.info(f"Generation model config: {self.cfg}")
        self.rank = int(os.getenv("RANK", 0))
        local_rank = int(os.getenv("LOCAL_RANK", 0))
        self.device = local_rank
        config = self.cfg

        ## 加载模型
        self.filelogger.info(f'loading {self.base_model}')
        self.model = WanModel.from_pretrained(self.base_model)
        self.model.eval().requires_grad_(False)

        # ## 初始化 T5 编码器
        self.shard_fn = partial(shard_model, device_id=self.device)
        self.text_encoder = T5EncoderModel(
            text_len=config.text_len,
            dtype=config.t5_dtype,
            device=torch.device('cpu'),
            checkpoint_path=os.path.join(self.base_model, config.t5_checkpoint),
            tokenizer_path=os.path.join(self.base_model, config.t5_tokenizer),
            shard_fn=self.shard_fn if self.t5_fsdp else None)

        ## 初始化 VAE（变分自编码器）
        self.vae = WanVAE(
            vae_pth=os.path.join(self.base_model, config.vae_checkpoint),
            device=self.device)
        
        if "t2v" in self.task:
            self.filelogger.info("Creating WanT2V pipeline.")
            self.wan_t2v = wan.WanT2V(
                config=self.cfg,
                checkpoint_dir=self.base_model,
                device_id=self.device,
                rank=self.rank,
                t5_fsdp=self.t5_fsdp,
                dit_fsdp=self.dit_fsdp,
                use_usp=(self.ulysses_size > 1 or self.ring_size > 1),
                t5_cpu=self.t5_cpu,
                model=self.model,
                vae=self.vae,
                shard_fn=self.shard_fn,
                text_encoder=self.text_encoder
            )

        if "i2v" in self.task:
            ### 初始化 CLIP 模型
            self.clip = CLIPModel(
                dtype=config.clip_dtype,
                device=self.device,
                checkpoint_path=os.path.join(self.base_model,config.clip_checkpoint),
                tokenizer_path=os.path.join(self.base_model, config.clip_tokenizer)
            )

            self.filelogger.info("Creating WanI2V pipeline.")
            self.wan_i2v = wan.WanI2V(
                config=self.cfg,
                checkpoint_dir=self.base_model,
                device_id=self.device,
                rank=self.rank,
                t5_fsdp=self.t5_fsdp,
                dit_fsdp=self.dit_fsdp,
                use_usp=(self.ulysses_size > 1 or self.ring_size > 1),
                t5_cpu=self.t5_cpu,
                model=self.model,
                vae=self.vae,
                shard_fn=self.shard_fn,
                text_encoder=self.text_encoder,
                clip=self.clip
            )
        self.filelogger.info("wrapperInit end")
        return 0


    def wrapperOnceExecAsync(self, params: {}, reqData: DataListCls, usrTag: str = "9527", persId: int = 0):
        res = self._wrapperOnceExec(params, reqData)
        r = callback(res, usrTag)
        if r != 0:
            self.filelogger.error(f'wrapperOnceExecAsync callback failed, ret code:{r}')
            return r
        return 0

    def wrapperOnceExec(self, params: {}, reqData: DataListCls, usrTag: str = "", persId: int = 0) -> Response:
        res = self._wrapperOnceExec(params, reqData)
        return res
    
    def _wrapperOnceExec(self, params: {}, reqData: DataListCls):
        self.filelogger.info(f"got params: {params}")
        prompts = str(self.get_payload_text(reqData))
        self.filelogger.info(f"Prompts: {prompts}")    
        w = int(params.get('width', 832))
        h = int(params.get('height', 480))
        self.size = f"{w}*{h}"
        if self.size not in SUPPORTED_SIZES[self.task]:
            self.filelogger.warning(f"Unsupport size {self.size}, use 832*480 instead")
            self.size = "832*480"
        self.sample_steps = int(params.get('sample_steps', 50))
        self.frame_num = int(params.get('frame_num', 81))
        
        try:
            if "t2v" in self.task:
                self.filelogger.info(f"Generating video ...")
                video = self.wan_t2v.generate(
                    prompts,
                    size=SIZE_CONFIGS[self.size],
                    frame_num=self.frame_num,
                    shift=self.sample_shift,
                    sample_solver=self.sample_solver,
                    sampling_steps=self.sample_steps,
                    guide_scale=self.sample_guide_scale,
                    seed=self.base_seed,
                    offload_model=self.offload_model
                )

            elif 'i2v' in self.task:
                self.sample_steps = 40
                self.filelogger.info("Set sample_steps to 40")
                self.image = self.get_image(reqData)
                if self.image is None:
                    self.filelogger.error("Failed to get image, video generate failed.")
                    return Response()
                if self.size in ["832*480", "480*832"]:
                    self.sample_shift = 3.0

                self.filelogger.info("Generating video ...")
                video = self.wan_i2v.generate(
                    prompts,
                    self.image,
                    max_area=MAX_AREA_CONFIGS[self.size],
                    frame_num=self.frame_num,
                    shift=self.sample_shift,
                    sample_solver=self.sample_solver,
                    sampling_steps=self.sample_steps,
                    guide_scale=self.sample_guide_scale,
                    seed=self.base_seed,
                    offload_model=self.offload_model
                )
            else:
                self.filelogger.error(f"Unsupported task type: {self.task}")
                return None
                
            video_bytes = self.video_to_bytes(
                tensor=video[None],
                fps=self.cfg.sample_fps,
                nrow=1,
                normalize=True,
                value_range=(-1, 1)
            )
            
            if video_bytes is None:
                self.filelogger.error("Failed to load video, video generate failed.")
                return None
            self.filelogger.info("Video generate successed.")

            # 封装返回值
            res = Response()
            resd = ResponseData()
            resd.key = "video"
            resd.setDataType(DataVideo)
            resd.status = Once
            resd.setData(video_bytes)
            res.list = [resd]
            return res
        
        except Exception as e:
            self.filelogger.error(f"Exception in _wrapperOnceExec: {str(e)}")
        return None
    
    def video_to_bytes(self, tensor, fps=30, suffix='.mp4', nrow=8, normalize=True, value_range=(-1, 1), retry=5):
        # 内存缓存
        buffer = io.BytesIO()

        # 保存到缓存
        for _ in range(retry):
            try:
                # 1. 限制值范围
                tensor = tensor.clamp(min(value_range), max(value_range))

                # 2. 逐帧转换
                tensor = torch.stack([
                    torchvision.utils.make_grid(
                        u, nrow=nrow, normalize=normalize, value_range=value_range)
                    for u in tensor.unbind(2)  # 拆分成单帧
                ], dim=1).permute(1, 2, 3, 0)  # (T, H, W, C)

                # 3. 归一化并转换为 uint8
                tensor = (tensor * 255).type(torch.uint8).cpu()

                # 4. 用 imageio 写入 BytesIO
                with imageio.get_writer(buffer, format=suffix.lstrip('.'), fps=fps, codec='libx264', quality=8) as writer:
                    for frame in tensor.numpy():
                        writer.append_data(frame)

                return buffer.getvalue()
                
            except Exception as e:
                self.filelogger.error(f'video_to_bytes failed, error: {e}')
        return None

    def wrapperFini(cls) -> int:
        return 0

    def wrapperError(cls, ret: int) -> str:
        if ret == 100:
            return "user error defined here"
        return ""

    '''
        此函数保留测试用，不可删除
    '''

    def wrapperTestFunc(cls, data: [], respData: []):
        pass


if __name__ == '__main__':
    m = Wrapper()
    m.run()
